@@ -2,9 +2,12 @@ import pkgutil
 from pathlib import Path
 from typing import Any, Dict, List, Optional
 
+import semver
 import typer
 import yaml
 from alteia.core.errors import ResponseError
+from alteia.core.utils.typing import ResourceId
+from alteia.sdk import SDK
 from PyInquirer import Separator, Token, prompt, style_from_dict
 from tabulate import tabulate
 
@@ -160,7 +163,7 @@ class Deliverable(Input):
 
 class Parameter:
     def __init__(self, name: str, *, display_name: str, description: str,
-                 required: bool, scheme: Dict = None, default_value: Optional[Any]):
+                 required: bool, scheme: Dict, default_value: Optional[Any]):
         self.name = name
         self.display_name = display_name
         self.required = required
@@ -191,6 +194,7 @@ class Parameter:
 class Analytic:
     def __init__(self, name: str, *, display_name: str, description: str,
                  docker_image: str,
+                 version: str = None,
                  instance_type: Optional[str] = None,
                  volume_size: Optional[int] = None,
                  tags: Optional[List[str]] = None, groups: List[str] = None,
@@ -201,6 +205,7 @@ class Analytic:
         self.display_name = display_name
         self.description = description
         self.docker_image = docker_image
+        self.version = version
         self.instance_type = instance_type
         self.volume_size = volume_size
         self.tags = tags
@@ -212,11 +217,13 @@ class Analytic:
     @classmethod
     def from_yaml(cls, yaml_desc: Dict):
         creation_params: Dict[str, Any]
+        version = yaml_desc.get('version')
         creation_params = {
             'name': str(yaml_desc.get('name')),
             'display_name': str(yaml_desc.get('display-name')),
             'description': str(yaml_desc.get('description')),
             'docker_image': str(yaml_desc.get('docker-image')),
+            'version': str(version) if version else None,
             'volume_size': yaml_desc.get('volume-size'),
             'instance_type': yaml_desc.get('instance-type'),
             'tags': yaml_desc.get('tags'),
@@ -252,13 +259,22 @@ ANALYTIC_DESC_SCHEMA = _get_analytic_schema('share/analytic_schema.yaml')
 
 @app.command(name='list')
 def list_analytics(
-    limit: int = typer.Option(default=100, help='Max number of analytics returned')
+    limit: int = typer.Option(default=100, help='Max number of analytics returned'),
+    display_all: bool = typer.Option(
+        False, '--all', case_sensitive=False,
+        help='If set, display all kinds of analytics '
+             '(otherwise only custom analytics are displayed).'
+    ),
 ):
     """ List the analytics """
     sdk = alteia_sdk()
     with utils.spinner():
+        search_filter = {'is_backoffice': {'$ne': True}}
+        if not display_all:
+            search_filter['external'] = {'$eq': True}
+
         found_analytics = sdk.analytics.search(
-            filter={'is_backoffice': {'$ne': True}},
+            filter=search_filter,
             return_total=True,
             limit=limit,
             sort={'display_name': 1}
@@ -272,6 +288,7 @@ def list_analytics(
                 for r in results
             ],
             'Name': [r.name for r in results],
+            'Version': [r.version if hasattr(r, 'version') else '' for r in results],
             'Identifier': [r.id for r in results]
         }
         print(tabulate(
@@ -298,7 +315,7 @@ def create(
         exists=True,
         readable=True,
         help='Path of the Analytic description (YAML file)'),
-        company: str = typer.Option(default=None, help='Company identifier'),
+    company: str = typer.Option(default=None, help='Company identifier'),
 ):
     """ Create a new analytic """
     analytic_desc = parse_analytic_yaml(description)
@@ -307,27 +324,37 @@ def create(
 
     sdk = alteia_sdk()
 
-    if analytic.name and not company:
+    if not company:
         company_shortname = analytic.name.split('/')[0]
-        companies = sdk._providers['auth_api'].post(
-            '/search-companies', {
-                'filter': {'short_name': {'$eq': company_shortname}},
-                'limit': 1
-            })
-        if companies['total'] != 1:
+        found = sdk.companies.search(
+            filter={'short_name': {'$eq': company_shortname}},
+            limit=1,
+            return_total=True
+        )
+        if found.total != 1:
             typer.secho(
                 f'✖ Impossible to find company with shortname {company_shortname}:',
                 fg=typer.colors.RED
             )
             raise typer.Exit(1)
 
-        company = companies['results'][0]['_id']
+        company = found.results[0]._id
 
-    found_analytics = sdk.analytics.search(name=analytic.name)
-    if found_analytics:
+    if not semver.VersionInfo.isvalid(analytic.version):
         typer.secho(
-            '⚠ {} already exists on {}'.format(
-                analytic.name, sdk._connection._base_url
+            '✖ The version {} does not respect the SemVer format'.format(
+                analytic.version
+            ),
+            fg=typer.colors.RED
+        )
+        raise typer.Exit(2)
+
+    found_analytics = find_analytics(sdk, analytic.name, analytic.version,
+                                     raise_if_not_found=False)
+    if len(found_analytics) != 0:
+        typer.secho(
+            '⚠ {} {} already exists on {}'.format(
+                analytic.name, analytic.version, sdk._connection._base_url
             ),
             fg=typer.colors.YELLOW
         )
@@ -337,14 +364,15 @@ def create(
         sdk.analytics.delete(analytic=found_analytics[0].id)
     else:
         typer.secho(
-            '✓ No analytic with the name {!r} on {!r}'.format(
-                analytic.name, sdk._connection._base_url
+            '✓ No analytic with the name {!r} and version {!r} on {!r}'.format(
+                analytic.name, analytic.version, sdk._connection._base_url
             ),
             fg=typer.colors.GREEN
         )
 
     analytic_creation_params = {
         'name': analytic.name,
+        'version': analytic.version,
         'display_name': analytic.display_name,
         'description': analytic.description,
         'docker_image': analytic.docker_image,
@@ -391,6 +419,20 @@ def create(
 def parse_analytic_yaml(analytic_yaml_path: Path) -> Dict:
     with open(analytic_yaml_path) as f:
         analytic = yaml.load(f, Loader=yaml.Loader)
+
+    # Specific message in case no version in the definition
+    # as analytic versioning is a new feature
+    if 'version' not in analytic:
+        typer.secho(
+            '⚠ From now on, you must specify an analytic version '
+            f'in your description file "{analytic_yaml_path.name}".\n\n'
+            'For instance, you can add under the "display-name: ..." line:\n'
+            'version: 1.0.0\n\n'
+            'Version must follow the SemVer syntax. See https://semver.org/.',
+            fg=typer.colors.YELLOW,
+        )
+        raise typer.Exit(2)
+
     errors = utils.validate_against_schema(
         analytic,
         json_schema=ANALYTIC_DESC_SCHEMA
@@ -407,31 +449,111 @@ def parse_analytic_yaml(analytic_yaml_path: Path) -> Dict:
         return analytic
 
 
-def find_analytic(sdk, analytic_name):
+def find_analytics(sdk, analytic_name: str, version_range: str = None,
+                   raise_if_not_found: bool = True):
+    search_filter = {'name': {'$eq': analytic_name}}
+    if version_range:
+        search_filter['version'] = {'$satisfies': version_range}
+
     with utils.spinner():
-        found_analytics = sdk.analytics.search(
-            name=analytic_name,
-            return_total=True,
-            limit=10,
+        results = list(sdk.analytics.search_generator(
+            filter=search_filter,
+            fields={'exclude': [
+                'algorithm',
+                'inputs',
+                'outputs',
+                'deliverables']}))
+
+    if raise_if_not_found and len(results) == 0:
+        msg = '✖ Cannot find the analytic {!r} {!r}'\
+            if version_range else '✖ Cannot find the analytic {!r}'
+        typer.secho(msg.format(analytic_name, version_range),
+                    fg=typer.colors.RED
+                    )
+        raise typer.Exit(2)
+
+    return results
+
+
+def get_companies_hierarchy(sdk: SDK):
+    user = sdk.users.describe()
+    company_ids = [c['id'] for c in user.companies]
+    user_companies = [c.__dict__ for c in sdk.companies.describe(company_ids)]
+    companies_by_id = {c['_id']: c for c in user_companies}
+    hierarchised_companies = list()
+    # Build a hierarchy of company (company object with property children)
+    for c in user_companies:
+        root_company = companies_by_id.get(c.get('root_company'))
+        if root_company:
+            children = root_company.get('children')
+            if children is None:
+                children = list()
+                root_company['children'] = children
+            children.append(c)
+        else:
+            hierarchised_companies.append(c)
+    # Sort children by name
+    for hierarchised_company in hierarchised_companies:
+        if 'children' not in hierarchised_company:
+            continue
+        hierarchised_company['children'] = sorted(
+            hierarchised_company['children'],
+            key=lambda c: c['name'])
+
+    return (user_companies, hierarchised_companies, user)
+
+
+def get_root_company(sdk: SDK) -> ResourceId:
+    # Try to find a root company available to the user
+    user = sdk.users.describe()
+    company_ids = [c['id'] for c in user.companies]
+    user_companies = sdk.companies.describe(company_ids)
+    root_company = None
+    for c in user_companies:
+        if c.domain == user.domain and not getattr(c, 'root_company', None):
+            root_company = c
+            break
+    else:
+        typer.secho(
+            '✖ No sufficient permissions to access the root company '
+            'of your domain.',
+            fg=typer.colors.RED
         )
-        results = found_analytics.results
-
-    if found_analytics.total == 1:
-        return results[0]
-
-    typer.secho(
-        '✖ Cannot find the analytic {!r}'.format(analytic_name),
-        fg=typer.colors.RED
-    )
-    raise typer.Exit(2)
+        raise typer.Exit(2)
+    return root_company._id
 
 
 @app.command()
 def unshare(
         analytic_name: str = typer.Argument(...),
-        company: str = typer.Option(default=None, help='Company identifier')):
+        version: str = typer.Option(default=None,
+                                    help='Range of versions in SemVer format. '
+                                    'If not provided, '
+                                    'all the versions will be unshared.'),
+        company: str = typer.Option(
+            default=None,
+            help='Identifier of the company to unshare the analytic with.'),
+        domain: bool = typer.Option(
+            default=False,
+            help='''To unshare the analytic with the root company of your domain.
+
+This is equivalent to using the --company option providing
+the id of the root company.
+Note that if you specifically shared the analytic with a company
+of your domain, the analytic will still be shared with that company.'''
+        )):
     """ Unshare an analytic """
     sdk = alteia_sdk()
+    analytics = find_analytics(sdk, analytic_name=analytic_name,
+                               version_range=version)
+    for analytic in analytics:
+        if not analytic.external:
+            typer.secho(
+                '✖ Cannot unshare the non-custom analytic {!r} {!r}'
+                .format(analytic_name, analytic.version),
+                fg=typer.colors.RED
+            )
+            raise typer.Exit(2)
 
     def company_selection(analytic_current_companies):
         def validate_answer(answer):
@@ -439,9 +561,7 @@ def unshare(
                 return True
             return 'You must choose at least company.'
 
-        user = sdk._providers['auth_api'].post(
-            '/describe-user', {'with_company_names': True})
-        companies = user['companies']
+        companies, hierarchised_companies, _ = get_companies_hierarchy(sdk)
         questions = [
             {
                 'type': 'checkbox',
@@ -454,86 +574,119 @@ def unshare(
             }
         ]
 
-        number_of_comp_shared = 0
-        for comp in companies:
-            choice = {'name': comp['name'], 'value': comp['id']}
-            if comp['id'] not in analytic_current_companies:
-                number_of_comp_shared += 1
+        number_of_comp_unshared = 0
+
+        def create_company_choice(comp, is_child, is_last_child):
+            nonlocal number_of_comp_unshared
+            name = comp['name']
+            choice = {
+                'name': f"└─ {name}" if is_last_child
+                else f"├─ {name}" if is_child
+                else name,
+                'value': comp['_id']
+            }
+            no_versions_shared_with_company = True
+            all_versions_owned_by_company = True
+            for version in analytics:
+                if comp['_id'] in version.companies:
+                    no_versions_shared_with_company = False
+                if comp['_id'] != version.company:
+                    all_versions_owned_by_company = False
+            if no_versions_shared_with_company or all_versions_owned_by_company:
+                number_of_comp_unshared += 1
                 choice['checked'] = True
-                choice['disabled'] = 'Analytics is not on this company'
+                choice['disabled'] = 'owns the analytic'\
+                    if all_versions_owned_by_company\
+                    else 'already not shared'
             questions[0]['choices'].append(choice)
 
-        if number_of_comp_shared == len(companies):
-            typer.secho(
-                '✖ Analytics isn\'t on any company',
-                fg=typer.colors.YELLOW
-            )
-            raise typer.Exit(2)
-
-        form_answers = prompt(questions, style=style)
-        companies_ids = form_answers['form']
+        for comp in hierarchised_companies:
+            create_company_choice(comp, False, False)
+            if 'children' in comp:
+                children = comp['children']
+                for i, child in enumerate(children):
+                    create_company_choice(child, True, i == len(children)-1)
 
+        if number_of_comp_unshared == len(companies):
+            companies_ids = []
+        else:
+            form_answers = prompt(questions, style=style)
+            # if user cancels, form_answers.get('form') will be None
+            companies_ids = form_answers.get('form')
         return companies_ids
 
-    analytic = find_analytic(sdk, analytic_name=analytic_name)
-    if not analytic.external:
-        typer.secho(
-            '✖ Cannot unshare the non-external analytic {!r}'.format(analytic_name),
-            fg=typer.colors.RED
-        )
-        raise typer.Exit(2)
-    try:
-        analytic_id = analytic.id
-        if company:
-            sdk.analytics.unshare_with_company(
-                analytic=analytic_id,
-                company=company)
-        else:
-            companies = company_selection(analytic.companies)
-            if not companies:
-                typer.secho(
-                    '✖ You should at least unshare '
-                    'the analytic {!r} with one company'.format(analytic_name),
-                    fg=typer.colors.YELLOW
-                )
-                raise typer.Exit(1)
+    if company:
+        companies = [company]
+    elif domain:
+        companies = [get_root_company(sdk)]
+    else:
+        companies = company_selection(analytics)
+        if companies is None:
+            # User cancelled
+            return
 
-            for company in companies:
+    for analytic in analytics:
+        companies_to_unshare_with = [c for c in companies if c in analytic.companies]
+        try:
+            for company in companies_to_unshare_with:
                 sdk.analytics.unshare_with_company(
-                    analytic=analytic_id,
+                    analytic=analytic.id,
                     company=company)
+        except ResponseError as e:
+            typer.secho(
+                '✖ Cannot unshare the analytic {!r} {!r}'
+                .format(analytic_name, analytic.version),
+                fg=typer.colors.RED
+            )
+            typer.secho('details: {}'.format(str(e)), fg=typer.colors.RED)
+            raise typer.Exit(2)
 
-    except ResponseError as e:
-        typer.secho(
-            '✖ Cannot unshare the analytic {!r}'.format(analytic_name),
-            fg=typer.colors.RED
-        )
-        typer.secho('details: {}'.format(str(e)), fg=typer.colors.RED)
-        raise typer.Exit(2)
-
-    typer.secho(
-        '✓ Analytic {!r} unshared successfully'.format(analytic_name),
-        fg=typer.colors.GREEN
-    )
+        if len(companies_to_unshare_with) > 0:
+            typer.secho(
+                '✓ Analytic {!r} {!r} unshared successfully'
+                .format(analytic_name, analytic.version),
+                fg=typer.colors.GREEN
+            )
+        else:
+            typer.secho(
+                'ℹ No companies to unshare the analytic {!r} {!r} with'.format(
+                    analytic_name, analytic.version),
+                fg=typer.colors.BLUE
+            )
 
 
 @app.command()
 def share(
         analytic_name: str = typer.Argument(...),
-        company: str = typer.Option(default=None, help='Company identifier')):
+        version: str = typer.Option(default=None,
+                                    help='Range of versions in SemVer format. '
+                                    'If not provided, '
+                                    'all the versions will be shared.'),
+        company: str = typer.Option(
+            default=None,
+            help='''Identifier of the company to share the analytic with.
+
+When providing the identifier of the root company of your domain,
+the analytic is shared with all the companies of the domain
+(equivalent to using the --domain option)'''),
+        domain: bool = typer.Option(
+            default=False,
+            help='''To share the analytic with the root company of your domain.
+
+This has the effect to share the analytic with all the
+companies of your domain and is equivalent to using the
+--company option providing the id of the root company.'''
+        )):
     """ Share an analytic """
     sdk = alteia_sdk()
 
-    def company_selection(analytic_current_companies):
+    def company_selection(analytics):
         def validate_answer(answer):
             if len(answer):
                 return True
             return 'You must choose at least one company.'
 
-        user = sdk._providers['auth_api'].post(
-            '/describe-user', {'with_company_names': True})
-        companies = user['companies']
-        print([c['name'] for c in companies])
+        companies, hierarchised_companies, _ = get_companies_hierarchy(sdk)
         questions = [
             {
                 'type': 'checkbox',
@@ -547,95 +700,134 @@ def share(
         ]
 
         number_of_comp_shared = 0
-        for comp in companies:
-            choice = {'name': comp['name'], 'value': comp['id']}
-            if comp['id'] in analytic_current_companies:
+
+        def create_company_choice(comp, is_child, is_last_child,
+                                  all_versions_of_root_shared_with_company):
+            nonlocal number_of_comp_shared
+            name = comp['name']
+            choice = {
+                'name': f"└─ {name}" if is_last_child
+                else f"├─ {name}" if is_child
+                else f'{name}  (and all its attached companies)'
+                if comp.get('children')
+                else name,
+                'value': comp['_id']
+            }
+            all_versions_shared_with_company = True
+            if all_versions_of_root_shared_with_company is False:
+                for version in analytics:
+                    if comp['_id'] not in version.companies:
+                        all_versions_shared_with_company = False
+                        break
+            if all_versions_shared_with_company:
                 number_of_comp_shared += 1
                 choice['checked'] = True
-                choice['disabled'] = 'This analytic is already shared with this company'
+                choice['disabled'] = 'already shared'
             questions[0]['choices'].append(choice)
+            return all_versions_shared_with_company
 
+        for comp in hierarchised_companies:
+            root_is_shared = create_company_choice(comp, False, False, False)
+            if 'children' in comp:
+                children = comp['children']
+                for i, child in enumerate(children):
+                    create_company_choice(child, True, i == len(children)-1,
+                                          root_is_shared)
         if number_of_comp_shared == len(companies):
+            companies_ids = []
+        else:
+            form_answers = prompt(questions, style=style)
+            # if user cancels, form_answers.get('form') will be None
+            companies_ids = form_answers.get('form')
+        return companies_ids
+
+    analytics = find_analytics(sdk, analytic_name=analytic_name,
+                               version_range=version)
+    for analytic in analytics:
+        if not analytic.external:
             typer.secho(
-                '✖ This analytic is already shared with all the companies',
-                fg=typer.colors.YELLOW
+                '✖ Cannot share the non-custom analytic {!r} {!r}'
+                .format(analytic_name, analytic.version),
+                fg=typer.colors.RED
             )
             raise typer.Exit(2)
 
-        form_answers = prompt(questions, style=style)
-        companies_ids = form_answers['form']
+    if company:
+        companies = [company]
+    elif domain:
+        companies = [get_root_company(sdk)]
+    else:
+        companies = company_selection(analytics)
+        if companies is None:
+            # User cancelled
+            return
 
-        return companies_ids
-
-    analytic = find_analytic(sdk, analytic_name=analytic_name)
-    if not analytic.external:
-        typer.secho(
-            '✖ Cannot share the non-external analytic {!r}'.format(analytic_name),
-            fg=typer.colors.RED
-        )
-        raise typer.Exit(2)
-    try:
-        analytic_id = analytic.id
-        if company:
-            sdk.analytics.share_with_company(
-                analytic=analytic_id,
-                company=company)
-        else:
-            companies = company_selection(analytic.companies)
-            if not companies:
-                typer.secho(
-                    '✖ You should at least share '
-                    'the analytic {!r} with one company'.format(analytic_name),
-                    fg=typer.colors.YELLOW
-                )
-                raise typer.Exit(1)
-
-            for company in companies:
+    for analytic in analytics:
+        companies_to_share_with = [c for c in companies if c not in analytic.companies]
+        try:
+            for company in companies_to_share_with:
                 sdk.analytics.share_with_company(
-                    analytic=analytic_id,
+                    analytic=analytic.id,
                     company=company)
+        except ResponseError as e:
+            typer.secho(
+                '✖ Cannot share the analytic {!r} {!r}'
+                .format(analytic_name, analytic.version),
+                fg=typer.colors.RED
+            )
+            typer.secho('details: {}'.format(str(e)), fg=typer.colors.RED)
+            raise typer.Exit(2)
 
-    except ResponseError as e:
-        typer.secho(
-            '✖ Cannot share the analytic {!r}'.format(analytic_name),
-            fg=typer.colors.RED
-        )
-        typer.secho('details: {}'.format(str(e)), fg=typer.colors.RED)
-        raise typer.Exit(2)
-
-    typer.secho(
-        '✓ Analytic {!r} shared successfully'.format(analytic_name),
-        fg=typer.colors.GREEN
-    )
+        if len(companies_to_share_with) > 0:
+            typer.secho(
+                '✓ Analytic {!r} {!r} shared successfully with {} companies'
+                .format(analytic_name, analytic.version,
+                        len(companies_to_share_with)),
+                fg=typer.colors.GREEN
+            )
+        else:
+            typer.secho(
+                'ℹ Analytic {!r} {!r} is already shared'.format(
+                    analytic_name, analytic.version),
+                fg=typer.colors.BLUE
+            )
 
 
 @app.command()
-def delete(analytic_name: str = typer.Argument(...)):
+def delete(analytic_name: str = typer.Argument(...),
+           version: str = typer.Option(default=None,
+                                       help='Version range of the analytic '
+                                       'in SemVer format. If not provided, '
+                                       'all the versions will be deleted.')):
     """ Delete an analytic """
     sdk = alteia_sdk()
-    analytic = find_analytic(sdk, analytic_name=analytic_name)
+    analytics = find_analytics(sdk, analytic_name=analytic_name,
+                               version_range=version)
 
-    if not analytic.external:
+    if len([a for a in analytics if a.external]) != len(analytics):
         deletion_confirm_msg = typer.style(
-            'Analytic {!r} is NOT an external analytic. '
+            'Analytic {!r} is NOT a custom analytic. '
             'Are you sure you want to delete it anyway?'.format(analytic_name),
             fg=typer.colors.YELLOW)
         typer.confirm(deletion_confirm_msg, abort=True)
-    try:
-        analytic_id = analytic.id
-        sdk.analytics.delete(analytic=analytic_id)
-    except ResponseError as e:
-        typer.secho(
-            '✖ Cannot delete the analytic {!r}'.format(analytic_name),
-            fg=typer.colors.RED
-        )
-        typer.secho('details: {}'.format(str(e)), fg=typer.colors.RED)
-        raise typer.Exit(2)
+    for analytic in analytics:
+        try:
+            analytic_id = analytic.id
+            sdk.analytics.delete(analytic=analytic_id)
+        except ResponseError as e:
+            typer.secho(
+                '✖ Cannot delete the analytic {!r} {!r}'
+                .format(analytic_name, analytic.version),
+                fg=typer.colors.RED
+            )
+            typer.secho('details: {}'.format(str(e)), fg=typer.colors.RED)
+            raise typer.Exit(2)
 
-    typer.secho(
-        '✓ Analytic {!r} deleted successfully'.format(analytic_name),
-        fg=typer.colors.GREEN
-    )
+        typer.secho(
+            '✓ Analytic {!r} {!r} deleted successfully'
+            .format(analytic_name, analytic.version),
+            fg=typer.colors.GREEN
+        )
 
 
 if __name__ == "__main__":
