@@ -6,7 +6,7 @@ from alteia_cli import config
 
 
 def alteia_sdk(credential_path: Path = config.DEFAULT_CREDENTIAL_PATH, *,
-               profile: str = config.DEFAULT_PROFILE):
+               profile: str = config.DEFAULT_PROFILE) -> alteia.SDK:
     credentials = config.get_credentials(
         credential_path=credential_path,
         profile=profile
